import sys

from potentials.record import Record, load_record, recordmanager
__all__ = ['Record', 'load_record', 'recordmanager']

#### Full record styles - include in recordmanager ####

# Import CrystalPrototype
try:
    from .CrystalPrototype import CrystalPrototype
except Exception as e:
    recordmanager.failed_styles['crystal_prototype'] = '%s: %s' % sys.exc_info()[:2]
else:
    recordmanager.loaded_styles['crystal_prototype'] = CrystalPrototype
    __all__.append('CrystalPrototype')

# Import RelaxedCrystal
try:
    from .RelaxedCrystal import RelaxedCrystal
except Exception as e:
    recordmanager.failed_styles['relaxed_crystal'] = '%s: %s' % sys.exc_info()[:2]
else:
    recordmanager.loaded_styles['relaxed_crystal'] = RelaxedCrystal
    __all__.append('RelaxedCrystal')

# Import ReferenceCrystal
try:
    from .ReferenceCrystal import ReferenceCrystal
except Exception as e:
    recordmanager.failed_styles['reference_crystal'] = '%s: %s' % sys.exc_info()[:2]
else:
    recordmanager.loaded_styles['reference_crystal'] = ReferenceCrystal
    __all__.append('ReferenceCrystal')

__all__.sort()


# coding: utf-8

# atomman imports
from ...library import Database

def load(id, api_key=None, database=None, local=None, remote=None,
         refresh_cache=False, verbose=False):
    """
    Loads a reference crystal structure from a DFT database.  Will first search
    the local database, then the remote NIST database to see if a matching
    entry is already included.  If a match is not found, then the structure
    will be obtained from the DFT database

    Parameters
    ----------
    id : str
        The reference crystal's unique id.  Combines a database tag "mp-" or
        "oqmd-" and the DFT database's entry id.
    api_key : str, optional
        The user's Materials Project API key or path to a file containing the
        key. Only needed for fetching structures from Materials Project and if
        the key is not set to the "MAPI_KEY" environment variable.
    database : atomman.library.Database, optional
        A pre-defined Database object to use.  If not given, will initialize
        a new Database object.  Passing in a database can save time if multiple
        calls are made for the same record type. 
    local : bool, optional
        Indicates if the Database object is to look for local records.  Default
        is True.  Ignored if database is given.
    remote : bool, optional
        Indicates if the Database object is to look for remote records.  Default
        is True.  Ignored if database is given.
    refresh_cache : bool, optional
        If the local database is of style "local", indicates if the metadata
        cache file is to be refreshed.  If False,
        metadata for new records will be added but the old record metadata
        fields will not be updated.  If True, then the metadata for all
        records will be regenerated, which is needed to update the metadata
        for modified records.
    verbose : bool, optional
        If True, info messages will be printed during operations.  Default
        value is False.
    
    Returns
    -------
    system : atomman.System
        The system object generated from the reference crystal.
    """
    # Create Database object and load if needed
    if database is None:
        database = Database.Database()
    
    # Fetch crystal from NIST or DFT database
    crystal = database.fetch_reference_crystal(id, api_key=api_key,
                                               local=local,
                                               remote=remote, 
                                               refresh_cache=refresh_cache,
                                               verbose=verbose)
    
    return crystal.ucell


# coding: utf-8

from . import Database
def load_lammps_potential(name=None, key=None, id=None,
                          potid=None, potkey=None, units=None,
                          atom_style=None, pair_style=None, status='active',
                          symbols=None, elements=None, pot_dir_style=None,
                          kim_models=None, kim_api_directory=None, kim_models_file=None, 
                          local=None, remote=None, database=None,
                          getfiles=False, prompt=True, verbose=False):
    """
    Loads a LAMMPS potential from the NIST Interatomic Potentials Repository
    or from a local copy of the repository.  Will issue a prompt if
    multiple LAMMPS potentials match the parameters given.
    
    Parameters
    ----------
    name : str or list
        The record name(s) to parse by.  For potential records, the names
        should correspond to the id with a prefix of "potentials." added to it.
    key : str or list
        The unique UUID4 record key(s) to parse by. 
    id : str or list
        The unique record id(s) labeling the records to parse by.
    potid : str or list
        The unique UUID4 record key(s) for the associated potential records to
        parse by.
    potkey : str or list
        The unique record id(s) labeling the associated potential records to
        parse by.
    units : str or list
        LAMMPS units option(s) to parse by.
    atom_style : str or list
        LAMMPS pair_style(s) to parse by.
    pair_style : str or list
        LAMMPS pair_style(s) to parse by.
    status : None, str or list
        Limits the search by the status of the LAMMPS implementations:
        "active", "superseded" and/or "retracted".  By default, only active
        implementations are returned.  Giving a value of None will return
        implementations of all statuses.
    symbols : str or list
        Model symbol(s) to parse by.  Typically correspond to elements for
        atomic potential models.
    elements : str or list
        Element(s) in the model to parse by.
    pot_dir_style : str, optional
        Specifies how the pot_dir values will be set for the retrieved LAMMPS
        potentials.  Allowed values are 'working', 'id', and 'local'.
        'working' will set all pot_dir = '', meaning parameter files
        are expected in the working directory when the potential is accessed.
        'id' sets the pot_dir values to match the potential's id.
        'local' sets the pot_dir values to the corresponding local database
        paths where the files are expected to be found.  Default value is
        controlled by settings.
    kim_models : list
        A list of full KIM model ids to build LAMMPS potentials for.
    kim_api_directory : str
        The path to the directory containing a kim-api-collections-management
        executable to use to identify which KIM models are installed.
    kim_models_file : str
        The path to a file containing a list of full KIM model ids to build
        LAMMPS potentials for.
    local : bool, optional
        Indicates if the local location is to be searched.  Default value
        matches the value set when the database was initialized.
    remote : bool, optional
        Indicates if the remote location is to be searched.  Default value
        matches the value set when the database was initialized.
    database : potentials.Database, optional
        Allows for a previously defined Database object to be used to find
        the potential.  If not given, a new Database object will be used with
        the default local and remote interaction settings.
    getfiles : bool, optional
        If True, then the parameter files for the matching potentials
        will also be copied/downloaded to the potential directory.
    prompt : bool
        If prompt=True (default) then a screen input will ask for a selection
        if multiple matching potentials are found.  If prompt=False, then an
        error will be thrown if multiple matches are found.
    verbose : bool, optional
        If True, info messages will be printed during operations.  Default
        value is False.
    
    Returns
    -------
    potentials.record.PotentialLAMMPS
        The potential object to use.

    Raises
    ------
    ValueError
        If no matching LAMMPS potentials are found.
    """
    # Create Database object and load if needed
    if database is None:
        database = Database()
    
    lmppot = database.get_lammps_potential(local=local, remote=remote, 
                                           name=name, key=key, id=id,
                                           potid=potid, potkey=potkey,
                                           units=units, atom_style=atom_style,
                                           pair_style=pair_style, status=status,
                                           symbols=symbols, elements=elements,
                                           pot_dir_style=pot_dir_style,
                                           kim_models=kim_models,
                                           kim_api_directory=kim_api_directory,
                                           kim_models_file=kim_models_file, 
                                           prompt=prompt, verbose=verbose)

    if getfiles is True:
        database.get_lammps_potential_files(lmppot, local=local, remote=remote,
                                            verbose=verbose)
    return lmppot


# coding: utf-8

# atomman imports
from . import run

def checkversion(lammps_command):
    """
    Gets the LAMMPS version and date information by testing lammps_command.
    
    Parameters
    ----------
    lammps_command: str
        A LAMMPS executable.
        
    Returns
    -------
    version_info: dict
        Dictionary containing 'version', the str LAMMPS version, and
        'date', the corresponding  datetime.date for the LAMMPS 
        version.
        
    Raises
    ------
    ValueError
        If lammps fails to run
    """
    # Run lammps_command with empty script and no log file
    try:
        log = run(lammps_command, script='', logfile=None)
    except:
        raise ValueError(f'Failed to run simulation with lammps_command {lammps_command}')
        
    # Extract lammps version and date info
    version_info = {}
    version_info['version'] = log.lammps_version
    version_info['date'] = log.lammps_date
    return version_info


# coding: utf-8

# atomman imports
from ...library import Database
from ...tools import aslist

import pandas as pd

def load(name=None, key=None, method='dynamic', standing='good',
         family=None, parent_key=None, potential=None,
         potential_LAMMPS_id=None, potential_LAMMPS_key=None, potential_id=None, potential_key=None,
         symbols=None, natoms=None, natypes=None, prompt=True, refresh_cache=False,
         verbose=False, database=None, local=False, remote=True):

    """
    Loads a potential-dependent relaxed crystal record from the library.  If
    multiple matches are found based on inputs a selection menu will appear.

    Parameters
    ----------
    name : str or list
        The record name(s) to parse by.  For relaxed crystal records, the
        names should correspond to the key.
    key : str or list, optional
        UUID4 key(s) to search for.  Each entry has a unique random-generated
        UUID4 key.
    method : str or list or None, optional
        The relaxation method used.  Allowed values are dynamic, static and box.
        Default value is dynamic (the most rigorous relaxation method).  All
        will be loaded if set to None.
    standing : str or list or None, optional
        "good" records are the unique crystals found with the most rigorous
        relaxation, and with known prototypes over DFT structures.  "bad" are
        records filtered out, usually for being duplicates.  Default value is
        "good".  All will be loaded if set to None.
    family : str or atomman.library.CrystalPrototype or list, optional
        The crystal family associated with the relaxed crystal - either crystal
        prototype name or MP/OQMD database entry name.
    parent_key : str or list, optional
        The UUID4 key(s) assigned to the calculation that the record is based
        on.
    potential : atomman.lammps.Potential or list, optional
        A loaded LAMMPS potential object to limit the search by.
    potential_LAMMPS_id : str or list, optional
        The id for a LAMMPS implemented potential to limit the search by.
    potential_LAMMPS_key : str or list, optional
        The UUID4 for a LAMMPS implemented potential to limit the search by.
    potential_id : str or list, optional
        The id for a potential to limit the search by.
    potential_key : str or list, optional
        The UUID4 for a potential to limit the search by.
    symbols : str or list, optional
        Element symbols in the crystal to limit the search by.
    natypes : int or list, optional
        The number of unique element model symbols in the crystal to limit
        the search by.
    natoms : int or list, optional
        The number of unique atoms in the crystal's unit cell to limit the
        search by.
    database : atomman.library.Database, optional
        A pre-defined Database object to use.  If not given, will initialize
        a new Database object.  Passing in a database can save time if multiple
        calls are made for the same record type. 
    local : bool, optional
        Indicates if the Database object is to look for local records.  Default
        is True.  Ignored if database is given.
    remote : bool, optional
        Indicates if the Database object is to look for remote records.  Default
        is True.  Ignored if database is given.
    prompt : bool
        If prompt=True (default) then a screen input will ask for a selection
        if multiple matching potentials are found.  If prompt=False, then an
        error will be thrown if multiple matches are found.
    refresh_cache : bool, optional
        If the local database is of style "local", indicates if the metadata
        cache file is to be refreshed.  If False,
        metadata for new records will be added but the old record metadata
        fields will not be updated.  If True, then the metadata for all
        records will be regenerated, which is needed to update the metadata
        for modified records.
    verbose : bool, optional
        If True, info messages will be printed during operations.  Default
        value is False.
    
    Returns
    -------
    system : atomman.System
        The system object generated from the relaxed crystal.
    """
    # Create Database object and load if needed
    if database is None:
        database = Database.Database()
    
    if potential is not None:
        try:
            assert potential_LAMMPS_id is None
            assert potential_LAMMPS_key is None
            assert potential_id is None
            assert potential_key is None
        except:
            raise ValueError('potential cannot be given with the other potential parameters')
        potential_LAMMPS_key = []
        for pot in aslist(potential):
            potential_LAMMPS_key.append(pot.key)

    # Get crystal prototype record
    crystal = database.get_relaxed_crystal(local=local, remote=remote, name=name,
                           key=key, method=method, standing=standing, family=family,
                           parent_key=parent_key, potential_LAMMPS_id=potential_LAMMPS_id,
                           potential_LAMMPS_key=potential_LAMMPS_key,
                           potential_id=potential_id, potential_key=potential_key,
                           symbols=symbols, natoms=natoms, natypes=natypes,
                           prompt=prompt, refresh_cache=refresh_cache, verbose=verbose)
    
    # Retrieve unit cell information and set symbols
    return crystal.ucell


# coding: utf-8

def get_crystal_prototypes(self, local=None, remote=None, name=None, id=None,
                           key=None, commonname=None, prototype=None,
                           pearson=None, strukturbericht=None, sg_number=None,
                           sg_hm=None, sg_schoenflies=None, crystalfamily=None,
                           natypes=None, refresh_cache=False, return_df=False,
                           verbose=False):
    """
    Get all matching crystal prototypes from the database.
    
    Parameters
    ----------
    local : bool, optional
        Indicates if the local location is to be searched.  Default value
        matches the value set when the database was initialized.
    remote : bool, optional
        Indicates if the remote location is to be searched.  Default value
        matches the value set when the database was initialized.
    name : str or list
        The record name(s) to parse by.  For crystal prototype records, the
        names should correspond to the id.
    id : str or list, optional
        Prototype ID(s) to search for.  These are unique identifiers for each
        prototype based on comm.
    key : str or list, optional
        UUID4 key(s) to search for.  Each entry has a unique random-generated
        UUID4 key.
    commonname : str or list, optional
        Common name(s) to limit the search by.
    prototype : str or list, optional
        Prototype identifying composition(s) to limit the search by.
    pearson : str or list, optional
        The Pearson symbol(s) to limit the search by.
    strukturbericht : str or list, optional
        The strukturbericht identifier(s) to limit the search by.
    sg_number : int or list, optional
        The space group number(s) to limit the search by.
    sg_hm : str or list, optional
        The space group Hermann-Maguin identifier(s) to limit the search by.
    sg_schoenflies : str or list, optional
        The space group Schoenflies identifier(s) to limit the search by.
    crystalfamily : str, optional
        The crystal system family to limit the search by. 
    natypes : int, optional
        The number(s) of unique atom types to limit the search by.
    refresh_cache : bool, optional
        If the local database is of style "local", indicates if the metadata
        cache file is to be refreshed.  If False,
        metadata for new records will be added but the old record metadata
        fields will not be updated.  If True, then the metadata for all
        records will be regenerated, which is needed to update the metadata
        for modified records.
    return_df : bool, optional
        If True, then the corresponding pandas.Dataframe of metadata
        will also be returned.
    verbose : bool, optional
        If True, info messages will be printed during operations.  Default
        value is False.
    """

    return self.get_records('crystal_prototype', local=local, remote=remote, name=name,
                            id=id, key=key, commonname=commonname, prototype=prototype,
                            pearson=pearson, strukturbericht=strukturbericht,
                            sg_number=sg_number, sg_hm=sg_hm, sg_schoenflies=sg_schoenflies,
                            crystalfamily=crystalfamily, natypes=natypes,
                            refresh_cache=refresh_cache, 
                            return_df=return_df, verbose=verbose)    

def get_crystal_prototype(self, local=None, remote=None, name=None, id=None,
                          key=None, commonname=None, prototype=None,
                          pearson=None, strukturbericht=None, sg_number=None,
                          sg_hm=None, sg_schoenflies=None, crystalfamily=None,
                          natypes=None, prompt=True, refresh_cache=False, verbose=False):
    
    """
    Retrieves exactly one matching crystal prototype from the database.
    
    Parameters
    ----------
    local : bool, optional
        Indicates if the local location is to be searched.  Default value
        matches the value set when the database was initialized.
    remote : bool, optional
        Indicates if the remote location is to be searched.  Default value
        matches the value set when the database was initialized.
    name : str or list
        The record name(s) to parse by.  For crystal prototype records, the
        names should correspond to the id.
    id : str or list, optional
        Prototype ID(s) to search for.  These are unique identifiers for each
        prototype based on comm.
    key : str or list, optional
        UUID4 key(s) to search for.  Each entry has a unique random-generated
        UUID4 key.
    name : str or list, optional
        Common name(s) to limit the search by.
    prototype : str or list, optional
        Prototype identifying composition(s) to limit the search by.
    pearson : str or list, optional
        The Pearson symbol(s) to limit the search by.
    strukturbericht : str or list, optional
        The strukturbericht identifier(s) to limit the search by.
    sg_number : int or list, optional
        The space group number(s) to limit the search by.
    sg_hm : str or list, optional
        The space group Hermann-Maguin identifier(s) to limit the search by.
    sg_schoenflies : str or list, optional
        The space group Schoenflies identifier(s) to limit the search by.
    crystalfamily : str or list, optional
        The crystal system family to limit the search by. 
    natypes : int or list, optional
        The number(s) of unique atom types to limit the search by.
    prompt : bool
        If prompt=True (default) then a screen input will ask for a selection
        if multiple matching potentials are found.  If prompt=False, then an
        error will be thrown if multiple matches are found.
    refresh_cache : bool, optional
        If the local database is of style "local", indicates if the metadata
        cache file is to be refreshed.  If False,
        metadata for new records will be added but the old record metadata
        fields will not be updated.  If True, then the metadata for all
        records will be regenerated, which is needed to update the metadata
        for modified records.
    verbose : bool, optional
        If True, info messages will be printed during operations.  Default
        value is False.
    """
    return self.get_record('crystal_prototype', local=True, remote=False,
                           name=name, id=id, key=key, commonname=commonname,
                           prototype=prototype, pearson=pearson,
                           strukturbericht=strukturbericht, sg_number=sg_number,
                           sg_hm=sg_hm, sg_schoenflies=sg_schoenflies,
                           crystalfamily=crystalfamily, natypes=natypes,
                           prompt=prompt, refresh_cache=refresh_cache, verbose=verbose)
        

def download_crystal_prototypes(self, name=None, id=None, key=None,
                                commonname=None, prototype=None,
                                pearson=None, strukturbericht=None,
                                sg_number=None, sg_hm=None, sg_schoenflies=None,
                                crystalfamily=None, natypes=None,
                                overwrite=False, verbose=False):
    """
    Downloads crystal prototypes from the remote to the local.
    
    Parameters
    ----------
    name : str or list
        The record name(s) to parse by.  For crystal prototype records, the
        names should correspond to the id.
    id : str or list, optional
        Prototype ID(s) to search for.  These are unique identifiers for each
        prototype based on comm.
    key : str or list, optional
        UUID4 key(s) to search for.  Each entry has a unique random-generated
        UUID4 key.
    name : str or list, optional
        Common name(s) to limit the search by.
    prototype : str or list, optional
        Prototype identifying composition(s) to limit the search by.
    pearson : str or list, optional
        The Pearson symbol(s) to limit the search by.
    strukturbericht : str or list, optional
        The strukturbericht identifier(s) to limit the search by.
    sg_number : int or list, optional
        The space group number(s) to limit the search by.
    sg_hm : str or list, optional
        The space group Hermann-Maguin identifier(s) to limit the search by.
    sg_schoenflies : str or list, optional
        The space group Schoenflies identifier(s) to limit the search by.
    crystalfamily : str or list, optional
        The crystal system families to limit the search by. 
    natypes : int or list, optional
        The number(s) of unique atom types to limit the search by.
    overwrite : bool, optional
        Flag indicating if any existing local records with names matching
        remote records are updated (True) or left unchanged (False).  Default
        value is False.
    verbose : bool, optional
        If True, info messages will be printed during operations.  Default
        value is False.
    """

    self.download_records('crystal_prototype', name=name,
                          id=id, key=key, commonname=commonname, prototype=prototype,
                          pearson=pearson, strukturbericht=strukturbericht,
                          sg_number=sg_number, sg_hm=sg_hm, sg_schoenflies=sg_schoenflies,
                          crystalfamily=crystalfamily, natypes=natypes,
                          overwrite=overwrite, verbose=verbose)


# coding: utf-8

from DataModelDict import DataModelDict as DM

from potentials.record import PotentialLAMMPS, PotentialLAMMPSKIM

def Potential(model, name=None, pot_dir=None, kim_id=None):
    """
    Class initializer switch for PotentialLAMMPS or PotentialLAMMPSKIM objects.

    Parameters
    ----------
    model : str or file-like object
        A JSON/XML data model containing a potential-LAMMPS or a
        potential-LAMMPS-KIM branch.
    name : str, optional
        The record name to use.  If not given, this will be set to the
        potential's id.
    pot_dir : str, optional
        The path to a directory containing any artifacts associated with
        the potential.  Default value is None, which assumes any required
        files will be in the working directory when LAMMPS is executed.
    kim_id : str, optional
        The full KIM model id indicating the version of a KIM model to use.
        If not given, then the newest known version will be used.
    
    Returns
    -------
    potentials.record.PotentialLAMMPS
        Returned if model contains potential-LAMMPS content.
    potentials.record.PotentialLAMMPSKIM
        Returned if model contains potential-LAMMPS-KIM content.
    """
    model = DM(model)
    try:
        # Search for potential-LAMMPS branch
        model.find('potential-LAMMPS')
    except:
        try:
            # Search for potential-LAMMPS-KIM branch
            model.find('potential-LAMMPS-KIM')
        except:
            raise ValueError('Failed to find either potential-LAMMPS or potential-LAMMPS-KIM content')
        else:
            return PotentialLAMMPSKIM(model=model, name=name, id=kim_id)
    else:
        return PotentialLAMMPS(model=model, name=name, pot_dir=pot_dir)



# Import records and load local record styles
from .record import recordmanager, load_record

from .Database import Database
from .load_lammps_potential import load_lammps_potential

__all__ = ['Database', 'CrystalPrototype', 'RelaxedCrystal']


import potentials

from ...tools import aslist

class Database(potentials.Database):
    """
    Child of potentials.Database extended for interacting with structure
    and defect reference records.
    """
    # Class imports
    from ._crystal_prototype import (get_crystal_prototypes, get_crystal_prototype,
                                     download_crystal_prototypes)

    from ._relaxed_crystal import (get_relaxed_crystals, get_relaxed_crystal,
                                   download_relaxed_crystals)

    from ._reference_crystal import (get_reference_crystals, get_reference_crystal,
                                     download_reference_crystals, fetch_mp_crystal,
                                     fetch_oqmd_crystal, fetch_mp_crystals,
                                     fetch_reference_crystal)

    def download_all(self, status=None, downloadfiles=True, overwrite=False,
                     verbose=False):
        """
        Downloads potential and structure-related records from the remote
        location to the local location.

        Parameters
        ----------
        status : str, list or None, optional
            Only potential_LAMMPS records with the given status(es) will be
            downloaded.  Allowed values are 'active' , 'superseded', and 'retracted'.
            If None (default) is given, then all potentials will be downloaded.
        downloadfiles : bool, optional
            If True, the parameter files associated with the potential_LAMMPS
            record will also be downloaded.
        overwrite : bool, optional
            Flag indicating if any existing local records with names matching
            remote records are updated (True) or left unchanged (False).  Default
            value is False.
        verbose : bool, optional
            If True, info messages will be printed during operations.  Default
            value is False.
        """
        super().download_all(status=status, downloadfiles=downloadfiles,
                             overwrite=overwrite, verbose=verbose)

        self.download_crystal_prototypes(overwrite=overwrite, verbose=verbose)
        self.download_relaxed_crystals(overwrite=overwrite, verbose=verbose)
        #self.download_reference_crystals(overwrite=overwrite, verbose=verbose)


# coding: utf-8
from pathlib import Path
import requests

from ...load import load_pymatgen_Structure, load_poscar
from ..record import ReferenceCrystal
from ...tools import aslist

def get_reference_crystals(self, local=None, remote=None, name=None, key=None,
                           id=None, sourcename=None, sourcelink=None,
                           crystalfamily=None, composition=None,
                           symbols=None, natoms=None, natypes=None,
                           refresh_cache=False, return_df=False, verbose=False):
    """
    Get all matching reference crystals from the database.
    
    Parameters
    ----------
    local : bool, optional
        Indicates if the local location is to be searched.  Default value
        matches the value set when the database was initialized.
    remote : bool, optional
        Indicates if the remote location is to be searched.  Default value
        matches the value set when the database was initialized.
    name : str or list, optional
        The record name(s) to parse by.  For reference crystal records, the
        names should correspond to the id.
    id : str or list, optional
        The record id(s) to parse by.  For reference crystal records, the
        id are letters identifying the source database "mp-", "mvc-", or "oqmd-",
        followed by the source database's identification number.
    key : str or list, optional
        UUID4 key(s) to search for.  Each entry has a unique random-generated
        UUID4 key.
    sourcename : str or list, optional
        The full name of source DFT databases to limit the search by.
        "Materials Project" or "Open Quantum Materials Database".            
    sourcelink : str or list, optional
        The web link of the source DFT databases to limit the search by.
    crystalfamily : str or list, optional
        The crystal system families to limit the search by.
    composition : str or list, optional
        The reduced compositions of the structures to limit the search by.
        Element symbols are sorted alphabetically.
    symbols : str or list, optional
        Element symbols in the crystal to limit the search by.
    natoms : int or list, optional
        The number of unique atoms in the crystal's unit cell to limit the
        search by.
    natypes : int or list, optional
        The number(s) of unique atom types to limit the search by.
    refresh_cache : bool, optional
        If the local database is of style "local", indicates if the metadata
        cache file is to be refreshed.  If False,
        metadata for new records will be added but the old record metadata
        fields will not be updated.  If True, then the metadata for all
        records will be regenerated, which is needed to update the metadata
        for modified records.
    return_df : bool, optional
        If True, then the corresponding pandas.Dataframe of metadata
        will also be returned.
    verbose : bool, optional
        If True, info messages will be printed during operations.  Default
        value is False.
    """

    return self.get_records('reference_crystal', local=local, remote=remote, name=name,
                            key=key, id=id, sourcename=sourcename, sourcelink=sourcelink,
                            crystalfamily=crystalfamily, composition=composition,
                            symbols=symbols, natoms=natoms, natypes=natypes,
                            refresh_cache=refresh_cache, 
                            return_df=return_df, verbose=verbose)

def get_reference_crystal(self, local=None, remote=None, name=None,
                        key=None, id=None, sourcename=None, sourcelink=None,
                        crystalfamily=None, composition=None,
                        symbols=None, natoms=None, natypes=None,
                        prompt=True, refresh_cache=False, verbose=False):
    """
    Retrieves exactly one matching reference crystal from the database.
    
    Parameters
    ----------
    local : bool, optional
        Indicates if the local location is to be searched.  Default value
        matches the value set when the database was initialized.
    remote : bool, optional
        Indicates if the remote location is to be searched.  Default value
        matches the value set when the database was initialized.
    name : str or list, optional
        The record name(s) to parse by.  For reference crystal records, the
        names should correspond to the id.
    id : str or list, optional
        The record id(s) to parse by.  For reference crystal records, the
        id are letters identifying the source database "mp-", "mvc-", or "oqmd-",
        followed by the source database's identification number.
    key : str or list, optional
        UUID4 key(s) to search for.  Each entry has a unique random-generated
        UUID4 key.
    sourcename : str or list, optional
        The full name of source DFT databases to limit the search by.
        "Materials Project" or "Open Quantum Materials Database".            
    sourcelink : str or list, optional
        The web link of the source DFT databases to limit the search by.
    crystalfamily : str or list, optional
        The crystal system families to limit the search by.
    composition : str or list, optional
        The reduced compositions of the structures to limit the search by.
        Element symbols are sorted alphabetically.
    symbols : str or list, optional
        Element symbols in the crystal to limit the search by.
    natoms : int or list, optional
        The number of unique atoms in the crystal's unit cell to limit the
        search by.
    natypes : int or list, optional
        The number(s) of unique atom types to limit the search by.
    prompt : bool
        If prompt=True (default) then a screen input will ask for a selection
        if multiple matching potentials are found.  If prompt=False, then an
        error will be thrown if multiple matches are found.
    refresh_cache : bool, optional
        If the local database is of style "local", indicates if the metadata
        cache file is to be refreshed.  If False,
        metadata for new records will be added but the old record metadata
        fields will not be updated.  If True, then the metadata for all
        records will be regenerated, which is needed to update the metadata
        for modified records.
    verbose : bool, optional
        If True, info messages will be printed during operations.  Default
        value is False.
    """
    def promptfxn(df):
        header = '#  id           comp     c-family     natoms alat'
        print(header)
        
        js = df.sort_values(['composition', 'natoms', 'crystalfamily', 'a']).index
        for i, j in enumerate(js):
            crystal = df.loc[j]
            row =  f'{i+1:2} {crystal.id:12.12} '
            row += f'{crystal.composition:8.8} '
            row += f'{crystal.crystalfamily:12.12} '
            row += f'{crystal.natoms:5d} '
            row += f'{crystal.a:7.4f}'
            print(row)
        
        i = int(input('Please select one:')) - 1
        if i < 0 or i >= len(js):
            raise ValueError('Invalid selection')

        return js[i]

    return self.get_record('reference_crystal', local=local, remote=remote, name=name,
                           key=key, id=id, sourcename=sourcename, sourcelink=sourcelink,
                           crystalfamily=crystalfamily, composition=composition,
                           symbols=symbols, natoms=natoms, natypes=natypes,
                           prompt=prompt, promptfxn=promptfxn, refresh_cache=refresh_cache, 
                           verbose=verbose)

def download_reference_crystals(self, name=None, key=None, id=None,
                                sourcename=None, sourcelink=None,
                                crystalfamily=None, composition=None,
                                symbols=None, natoms=None, natypes=None, keyword=None,
                                overwrite=False, verbose=False):
    """
    Download reference records from the remote and save to localpath.
    
    Parameters
    ----------
    name : str or list, optional
        The record name(s) to parse by.  For reference crystal records, the
        names should correspond to the id.
    id : str or list, optional
        The record id(s) to parse by.  For reference crystal records, the
        id are letters identifying the source database "mp-", "mvc-", or "oqmd-",
        followed by the source database's identification number.
    key : str or list, optional
        UUID4 key(s) to search for.  Each entry has a unique random-generated
        UUID4 key.
    sourcename : str or list, optional
        The full name of source DFT databases to limit the search by.
        "Materials Project" or "Open Quantum Materials Database".            
    sourcelink : str or list, optional
        The web link of the source DFT databases to limit the search by.
    crystalfamily : str or list, optional
        The crystal system families to limit the search by.
    composition : str or list, optional
        The reduced compositions of the structures to limit the search by.
        Element symbols are sorted alphabetically.
    symbols : str or list, optional
        Element symbols in the crystal to limit the search by.
    natoms : int or list, optional
        The number of unique atoms in the crystal's unit cell to limit the
        search by.
    natypes : int or list, optional
        The number(s) of unique atom types to limit the search by.
    overwrite : bool, optional
        Flag indicating if any existing local records with names matching
        remote records are updated (True) or left unchanged (False).  Default
        value is False.
    verbose : bool, optional
        If True, info messages will be printed during operations.  Default
        value is False.
    """

    self.download_records('relaxed_crystal', name=name,
                          key=key, id=id, sourcename=sourcename, sourcelink=sourcelink,
                          crystalfamily=crystalfamily, composition=composition,
                          symbols=symbols, natoms=natoms, natypes=natypes,
                          overwrite=overwrite, verbose=verbose)

def fetch_reference_crystal(self, id, api_key=None, local=None, remote=None, 
                            refresh_cache=False, verbose=False):
    """
    Retrieves a single reference crystal.  First, the database is checked
    for matches with the DOI, then with the record name.  If no matches are found
    in the database, then the corresponding crystal structure is downloaded from
    the source database.

    Parameters
    ----------
    id : str
        The reference crystal's unique id.  Combines a database tag "mp-" or
        "oqmd-" and the DFT database's entry id.
    api_key : str, optional
        The user's Materials Project API key or path to a file containing the
        key. Only needed for fetching structures from Materials Project and if
        the key is not set to the "MAPI_KEY" environment variable.
    local : bool, optional
        Indicates if the local location is to be searched.  Default value
        matches the value set when the database was initialized.
    remote : bool, optional
        Indicates if the remote location is to be searched.  Default value
        matches the value set when the database was initialized.
    refresh_cache : bool, optional
        If the local database is of style "local", indicates if the metadata
        cache file is to be refreshed.  If False,
        metadata for new records will be added but the old record metadata
        fields will not be updated.  If True, then the metadata for all
        records will be regenerated, which is needed to update the metadata
        for modified records.
    verbose : bool, optional
        If True, info messages will be printed during operations.  Default
        value is False.
    """
    if local is not False or remote is not False:
        # Try fetching based on doi
        try:
            return self.get_reference_crystal(id=id, local=local, remote=remote,
                                              refresh_cache=refresh_cache, verbose=verbose)
        except:
            pass

    # Fetch from source database
    if 'oqmd-' in id:
        record = self.fetch_oqmd_crystal(id)
        if verbose:
            print(f'Crystal retrieved from OQMD')
    else:
        record = self.fetch_mp_crystal(id, api_key=api_key)
        if verbose:
            print(f'Crystal retrieved from Materials Project')

    return record

def fetch_mp_crystals(self, id, api_key=None):
    """
    Retrieves reference crystals from Materials Project based on id(s).

    Parameters
    ----------
    id : str or list
        The structure id(s) of the crystals to retrieve.
    api_key : str, optional
        The user's Materials Project API key or path to a file containing the key.
        If not given, will use the "MAPI_KEY" environment variable.
    """

    # Function-specific imports
    import pymatgen as pmg
    from pymatgen.ext.matproj import MPRester

    # Open connection to Materials Project
    records = []
    with MPRester(api_key) as m:

        # Download missing entries
        try:
            entries = m.query({"material_id": {"$in": aslist(id)}}, ['material_id', 'cif'])
        except:
            raise ValueError('Failed to find Materials Project information')
        else:
            # Convert cif to model and save
            for entry in entries:
                entry_id = entry['material_id']
                struct = pmg.Structure.from_str(entry['cif'], fmt='cif')
                struct = pmg.symmetry.analyzer.SpacegroupAnalyzer(struct).get_conventional_standard_structure()

                # Build record content
                # Build basic record content
                record = ReferenceCrystal(name=entry_id)
                record.sourcename = "Materials Project"
                record.sourcelink = "https://materialsproject.org/"
                record.ucell = load_pymatgen_Structure(struct).normalize()
                records.append(record)
    
    return records

def fetch_mp_crystal(self, id, api_key=None):
    """
    Retrieves a single reference crystal from Materials Project based on id.

    Parameters
    ----------
    id : str
        The structure id of the crystal to retrieve.
    api_key : str, optional
        The user's Materials Project API key. If not given, will use "MAPI_KEY"
        environment variable.
    """
    # Read api_key from a file
    if Path(api_key).is_file():
        with open(api_key) as f:
            api_key = f.read().strip()

    records = self.fetch_mp_crystals(id, api_key=api_key)
    if len(records) == 1:
        return records[0]
    else:
        raise ValueError('Exactly one record not found?')

def fetch_oqmd_crystal(self, id):
    """
    Retrieves a single reference crystal from OQMD based on id.

    Parameters
    ----------
    id : str
        The OQMD entry number with "oqmd-" prefix.
    """
    
    # Build basic record content
    record = ReferenceCrystal(name=id)
    record.sourcename = "Open Quantum Materials Database"
    record.sourcelink = "http://oqmd.org/"
    
    # Parse entry number page for structure number
    entry_number = id.replace('oqmd-', '')
    entry_r = requests.get(f'http://oqmd.org/materials/entry/{entry_number}')
    entry_html = entry_r.text        
    start = entry_html.index('href="/materials/structure/') + len('href="/materials/structure/')
    end = start + entry_html[start:].index('"')
    structure_number = entry_html[start:end]

    # Try retrieving poscar of conventional then primitive cells
    try:
        structure_url = f'http://oqmd.org/materials/export/conventional/poscar/{structure_number}'
        structure_r = requests.get(structure_url)
        structure_r.raise_for_status()
    except:
        try:
            structure_url = f'http://oqmd.org/materials/export/primitive/poscar/{structure_number}' 
            structure_r = requests.get(structure_url)
            structure_r.raise_for_status()
        except:
            raise ValueError('Failed to find the poscar file for the structure')
            
    # Load ucell
    record.ucell = load_poscar(structure_r.text).normalize()
    
    return record

def save_reference_crystal(self, crystal, overwrite=False, verbose=False):
    """
    Saves a citation to the local database.
    
    Parameters
    ----------
    crystal : ReferenceCrystal
        The record to save.  
    overwrite : bool, optional
        Indicates what to do when a matching record is found in the local
        location.  If False (default), then the record is not updated.  If
        True, then the record is updated.
    verbose : bool, optional
        If True, info messages will be printed during operations.  Default
        value is False.
    """
    self.save_record(record=crystal, overwrite=overwrite, verbose=verbose)


from copy import deepcopy
import uuid

# https://github.com/usnistgov/DataModelDict
from DataModelDict import DataModelDict as DM

from datamodelbase.record import Record
from datamodelbase import query

from ... import System
from ...tools import crystalsystem

modelroot = 'reference-crystal'

class ReferenceCrystal(Record):
    
    @property
    def style(self):
        """str: The record style"""
        return 'reference_crystal'

    @property
    def modelroot(self):
        """str: The root element of the content"""
        return modelroot

    @property
    def xsd_filename(self):
        return ('atomman.library.xsd', f'{self.style}.xsd')

    @property
    def id(self):
        """str : The unique id for the record"""
        return self.__id

    @id.setter
    def id(self, value):
        if value is None:
            self.__id = None
        else:
            self.__id = str(value)

    @property
    def key(self):
        """str : A UUID4 key assigned to the record"""
        return self.__key

    @key.setter
    def key(self, value):
        if value is None:
            self.__key = str(uuid.uuid4())
        else:
            self.__key = str(value)

    @property
    def sourcename(self):
        """str : Name of the crystal's source database"""
        return self.__sourcename

    @sourcename.setter
    def sourcename(self, value):
        if value is None:
            self.__sourcename = None
        else:
            self.__sourcename = str(value)

    @property
    def sourcelink(self):
        """str : URL for the crystal's source database"""
        return self.__sourcelink    

    @sourcelink.setter
    def sourcelink(self, value):
        if value is None:
            self.__sourcelink = None
        else:
            self.__sourcelink = str(value)

    @property
    def ucell(self):
        """atomman.System : The unit cell system for the crystal"""
        if self.__ucell is None:
            raise ValueError('ucell information not set')
        elif not isinstance(self.__ucell, System):
            self.__ucell = System(model=self.__ucell)
        return self.__ucell

    @ucell.setter
    def ucell(self, value):        
        self.__ucell = value

    @property
    def composition(self):
        """str : The crystal's composition"""
        if self.__composition is None:
            self.__composition = self.ucell.composition
        return self.__composition

    @property
    def symbols(self):
        """list : The list of element model symbols"""
        if self.__symbols is None:
            self.__symbols = self.ucell.symbols
        return self.__symbols

    @property
    def natoms(self):
        """int : The number of atoms in the unit cell"""
        if self.__natoms is None:
            self.__natoms = self.ucell.natoms
        return self.__natoms

    @property
    def natypes(self):
        """int : The number of unique elements in the unit cell"""
        if self.__natypes is None:
            self.__natypes = self.ucell.natypes
        return self.__natypes

    @property
    def crystalfamily(self):
        """str : The crystal's system family"""
        if self.__crystalfamily is None:
            self.__crystalfamily = crystalsystem.identifyfamily(self.ucell.box)
        return self.__crystalfamily

    @property
    def a(self):
        """float : The unit cell's a lattice parameter"""
        if self.__a is None:
            self.__a = self.ucell.box.a
        return self.__a

    @property
    def b(self):
        """float : The unit cell's b lattice parameter"""
        if self.__b is None:
            self.__b = self.ucell.box.b
        return self.__b

    @property
    def c(self):
        """float : The unit cell's c lattice parameter"""
        if self.__c is None:
            self.__c = self.ucell.box.c
        return self.__c

    @property
    def alpha(self):
        """float : The unit cell's alpha lattice angle"""
        if self.__alpha is None:
            self.__alpha = self.ucell.box.alpha
        return self.__alpha

    @property
    def beta(self):
        """float : The unit cell's beta lattice angle"""
        if self.__beta is None:
            self.__beta = self.ucell.box.beta
        return self.__beta

    @property
    def gamma(self):
        """float : The unit cell's gamma lattice angle"""
        if self.__gamma is None:
            self.__gamma = self.ucell.box.gamma
        return self.__gamma

    def set_values(self, name=None, id=None, key=None, sourcename=None,
                   sourcelink=None, ucell=None):
        """
        Sets multiple object values.

        Parameters
        ----------
        name : str, optional
            The name to use for saving the record.  Either name or id should
            be given as they are treated as aliases for this record style.
        id : str, optional
            The unique identifier for the record.  Should be composed of a
            source database tag plus the source database's unique identifier.
        key : str, optional
            A UUID4 key assigned to the record.  Note that if not given a new
            random key will be assigned and therefore the keys might not match
            if similar records were generated independently.
        sourcename : str, optional
            The name of the source database where the reference record was
            retrieved.
        sourcelink : str, optional
            The URL to the source database where the reference record was
            retrieved.
        ucell : atomman.System, optional
            A small unit cell system associated with the reference crystal.
        """
        
        if name is None and id is not None:
            self.name = id
            self.id = id
        elif name is not None and id is None:
            self.name = name
            self.id = name
        else:
            self.name = name
            self.id = id

        self.key = key
        self.sourcename = sourcename
        self.sourcelink = sourcelink
        self.ucell = ucell
        
        self.__symbols = None
        self.__composition = None
        self.__crystalfamily = None
        self.__natypes = None
        self.__a = None
        self.__b = None
        self.__c = None
        self.__alpha = None
        self.__beta = None
        self.__gamma = None

    def build_model(self):
        """
        Returns the object info as data model content
        
        Returns
        ----------
        DataModelDict
            The data model content.
        """

        refmodel = DM()
        
        refmodel['key'] = self.key
        refmodel['id'] = self.id

        refmodel['source'] = DM()
        refmodel['source']['name'] = self.sourcename
        refmodel['source']['link'] = self.sourcelink
        
        refmodel['system-info'] = DM()
        symbols = self.symbols
        if len(symbols) == 1:
            refmodel['system-info']['symbol'] = self.symbols[0]
        else:
            refmodel['system-info']['symbol'] = self.symbols
        refmodel['system-info']['composition'] = self.composition
        
        refmodel['system-info']['cell'] = DM()
        refmodel['system-info']['cell']['crystal-family'] = self.crystalfamily
        refmodel['system-info']['cell']['natypes'] = self.natypes
        refmodel['system-info']['cell']['a'] = self.a
        refmodel['system-info']['cell']['b'] = self.b
        refmodel['system-info']['cell']['c'] = self.c
        refmodel['system-info']['cell']['alpha'] = self.alpha
        refmodel['system-info']['cell']['beta'] = self.beta
        refmodel['system-info']['cell']['gamma'] = self.gamma
        
        refmodel['atomic-system'] = self.ucell.model()['atomic-system']

        model = DM([('reference-crystal', refmodel)])
        self._set_model(model)
        return model
        
    def load_model(self, model, name=None):
        super().load_model(model, name=name)        
        crystal = self.model[modelroot]
        
        self.__key = crystal['key']
        self.__id = crystal['id']
        self.__sourcename = crystal['source']['name']
        self.__sourcelink = crystal['source']['link']
        
        self.__symbols = crystal['system-info'].aslist('symbol')
        self.__composition = crystal['system-info']['composition']
        self.__crystalfamily = crystal['system-info']['cell']['crystal-family']
        self.__natypes = crystal['system-info']['cell']['natypes']
        self.__a = crystal['system-info']['cell']['a']
        self.__b = crystal['system-info']['cell']['b']
        self.__c = crystal['system-info']['cell']['c']
        self.__alpha = crystal['system-info']['cell']['alpha']
        self.__beta = crystal['system-info']['cell']['beta']
        self.__gamma = crystal['system-info']['cell']['gamma']

        self.__natoms = crystal['atomic-system']['atoms']['natoms']

        self.__ucell = crystal

        # Set name as id if no name given
        try:
            self.name
        except:
            self.name = self.id

    def metadata(self):
        """
        Converts the structured content to a simpler dictionary.
        
        Returns
        -------
        dict
            A dictionary representation of the record's content.
        """
        params = {}
        params['name'] = self.name
        params['key'] = self.key
        params['id'] = self.id
        params['sourcename'] = self.sourcename
        params['sourcelink'] = self.sourcelink

        params['crystalfamily'] = self.crystalfamily
        params['natypes'] = self.natypes
        params['symbols'] = self.symbols
        params['composition'] = self.composition

        params['a'] = self.a
        params['b'] = self.b
        params['c'] = self.c
        params['alpha'] = self.alpha
        params['beta'] = self.beta
        params['gamma'] = self.gamma
        params['natoms'] = self.natoms

        return params

    @staticmethod
    def pandasfilter(dataframe, name=None, key=None,
                     id=None, sourcename=None,
                     sourcelink=None, crystalfamily=None, composition=None,
                     symbols=None, natoms=None, natypes=None):

        matches = (
            query.str_match.pandas(dataframe, 'name', name)
            &query.str_match.pandas(dataframe, 'key', key)
            &query.str_match.pandas(dataframe, 'id', id)
            &query.str_match.pandas(dataframe, 'sourcename', sourcename)
            &query.str_match.pandas(dataframe, 'sourcelink', sourcelink)
            &query.str_match.pandas(dataframe, 'crystalfamily', crystalfamily)
            &query.str_match.pandas(dataframe, 'composition', composition)
            &query.in_list.pandas(dataframe, 'symbols', symbols)
            &query.str_match.pandas(dataframe, 'natoms', natoms)
            &query.str_match.pandas(dataframe, 'natypes', natypes)
        )
        return matches

    @staticmethod
    def mongoquery(name=None, key=None,
                   id=None, sourcename=None,
                   sourcelink=None, crystalfamily=None, composition=None,
                   symbols=None, natoms=None, natypes=None):

        mquery = {}
        query.str_match.mongo(mquery, f'name', name)
        root = f'content.{modelroot}'

        query.str_match.mongo(mquery, f'{root}.key', key)
        query.str_match.mongo(mquery, f'{root}.id', id)
        query.str_match.mongo(mquery, f'{root}.source.name', sourcename)
        query.str_match.mongo(mquery, f'{root}.sourcelink', sourcelink)
        query.str_match.mongo(mquery, f'{root}.system-info.cell.crystal-family', crystalfamily)
        query.str_match.mongo(mquery, f'{root}.system-info.cell.composition', composition)
        query.in_list.mongo(mquery, f'{root}.system-info.symbol', symbols)
        query.str_match.mongo(mquery, f'{root}.atomic-system.atoms.natoms', natoms)
        query.str_match.mongo(mquery, f'{root}.system-info.cell.natypes', natypes)

        return mquery

    @staticmethod
    def cdcsquery(key=None, id=None, sourcename=None,
                  sourcelink=None, crystalfamily=None, composition=None,
                  symbols=None, natoms=None, natypes=None):

        mquery = {}
        root = modelroot
        
        query.str_match.mongo(mquery, f'{root}.key', key)
        query.str_match.mongo(mquery, f'{root}.id', id)
        query.str_match.mongo(mquery, f'{root}.source.name', sourcename)
        query.str_match.mongo(mquery, f'{root}.sourcelink', sourcelink)
        query.str_match.mongo(mquery, f'{root}.system-info.cell.crystal-family', crystalfamily)
        query.str_match.mongo(mquery, f'{root}.system-info.cell.composition', composition)
        query.in_list.mongo(mquery, f'{root}.system-info.symbol', symbols)
        query.str_match.mongo(mquery, f'{root}.atomic-system.atoms.natoms', natoms)
        query.str_match.mongo(mquery, f'{root}.system-info.cell.natypes', natypes)

        return mquery


from copy import deepcopy

# https://github.com/usnistgov/DataModelDict
from DataModelDict import DataModelDict as DM

from datamodelbase.record import Record
from datamodelbase import query

import atomman.unitconvert as uc
from ... import System

modelroot = 'relaxed-crystal'

class RelaxedCrystal(Record):
    
    @property
    def style(self):
        """str: The record style"""
        return 'relaxed_crystal'

    @property
    def modelroot(self):
        """str: The root element of the content"""
        return modelroot

    @property
    def xsd_filename(self):
        return ('atomman.library.xsd', f'{self.style}.xsd')

    def load_model(self, model, name=None):
        super().load_model(model, name=name)        
        crystal = self.model[modelroot]
        
        self.__key = crystal['key']
        self.__method = crystal['method']
        self.__standing = crystal['standing']

        self.__potential_LAMMPS_id = crystal['potential-LAMMPS']['id']
        self.__potential_LAMMPS_key = crystal['potential-LAMMPS']['key']
        self.__potential_id = crystal['potential-LAMMPS']['potential']['id']
        self.__potential_key = crystal['potential-LAMMPS']['potential']['key']

        self.__family = crystal['system-info']['family']
        self.__parent_key = crystal['system-info']['parent_key']
        self.__symbols = crystal['system-info'].aslist('symbol')
        self.__composition = crystal['system-info']['composition']
        self.__crystalfamily = crystal['system-info']['cell']['crystal-family']
        self.__natypes = crystal['system-info']['cell']['natypes']
        self.__a = crystal['system-info']['cell']['a']
        self.__b = crystal['system-info']['cell']['b']
        self.__c = crystal['system-info']['cell']['c']
        self.__alpha = crystal['system-info']['cell']['alpha']
        self.__beta = crystal['system-info']['cell']['beta']
        self.__gamma = crystal['system-info']['cell']['gamma']
        
        self.__natoms = crystal['atomic-system']['atoms']['natoms']

        self.__potential_energy = uc.value_unit(crystal['potential-energy'])
        self.__cohesive_energy = uc.value_unit(crystal['cohesive-energy'])
        
        self.__ucell = None

        # Set name as key if no name given
        try:
            self.name
        except:
            self.name = self.key

    @property
    def key(self):
        return self.__key

    @property
    def method(self):
        return self.__method
    
    @property
    def standing(self):
        return self.__standing
    
    @property
    def family(self):
        return self.__family

    @property
    def parent_key(self):
        return self.__parent_key

    @property
    def potential_LAMMPS_id(self):
        return self.__potential_LAMMPS_id

    @property
    def potential_LAMMPS_key(self):
        return self.__potential_LAMMPS_key

    @property
    def potential_id(self):
        return self.__potential_id

    @property
    def potential_key(self):
        return self.__potential_key

    @property
    def cohesive_energy(self):
        return self.__cohesive_energy

    @property
    def potential_energy(self):
        return self.__potential_energy

    @property
    def composition(self):
        return self.__composition

    @property
    def symbols(self):
        return self.__symbols

    @property
    def natoms(self):
        return self.__natoms

    @property
    def natypes(self):
        return self.__natypes

    @property
    def crystalfamily(self):
        return self.__crystalfamily

    @property
    def a(self):
        return self.__a

    @property
    def b(self):
        return self.__b

    @property
    def c(self):
        return self.__c

    @property
    def alpha(self):
        return self.__alpha

    @property
    def beta(self):
        return self.__beta

    @property
    def gamma(self):
        return self.__gamma

    @property
    def ucell(self):
        if self.__ucell is None:
            self.__ucell = System(model=self.model)
        return self.__ucell

    def build_model(self):
        return self.model

    def metadata(self):
        """
        Converts the structured content to a simpler dictionary.
        
        Returns
        -------
        dict
            A dictionary representation of the record's content.
        """
        params = {}
        params['name'] = self.name
        params['key'] = self.key
        params['method'] = self.method
        params['standing'] = self.standing
        params['family'] = self.family
        params['parent_key'] = self.parent_key
        
        params['potential_LAMMPS_id'] = self.potential_LAMMPS_id
        params['potential_LAMMPS_key'] = self.potential_LAMMPS_key
        params['potential_id'] = self.potential_id
        params['potential_key'] = self.potential_key
        
        params['crystalfamily'] = self.crystalfamily
        params['natypes'] = self.natypes
        params['symbols'] = self.symbols
        params['composition'] = self.composition

        params['a'] = self.a
        params['b'] = self.b
        params['c'] = self.c
        params['alpha'] = self.alpha
        params['beta'] = self.beta
        params['gamma'] = self.gamma
        params['natoms'] = self.natoms

        params['potential_energy'] = self.potential_energy
        params['cohesive_energy'] = self.cohesive_energy

        return params

    @staticmethod
    def pandasfilter(dataframe, name=None, key=None,
                     method='dynamic', standing='good',
                     family=None, parent_key=None, 
                     potential_LAMMPS_id=None, potential_LAMMPS_key=None,
                     potential_id=None, potential_key=None,
                     crystalfamily=None, composition=None,
                     symbols=None, natoms=None, natypes=None):

        matches = (
            query.str_match.pandas(dataframe, 'name', name)
            &query.str_match.pandas(dataframe, 'key', key)
            &query.str_match.pandas(dataframe, 'method', method)
            &query.str_match.pandas(dataframe, 'standing', standing)
            &query.str_match.pandas(dataframe, 'family', family)
            &query.str_match.pandas(dataframe, 'parent_key', parent_key)
            &query.str_match.pandas(dataframe, 'potential_LAMMPS_id', potential_LAMMPS_id)
            &query.str_match.pandas(dataframe, 'potential_LAMMPS_key', potential_LAMMPS_key)
            &query.str_match.pandas(dataframe, 'potential_id', potential_id)
            &query.str_match.pandas(dataframe, 'potential_key', potential_key)
            &query.str_match.pandas(dataframe, 'potential_key', potential_key)
            &query.str_match.pandas(dataframe, 'crystalfamily', crystalfamily)
            &query.str_match.pandas(dataframe, 'composition', composition)
            &query.in_list.pandas(dataframe, 'symbols', symbols)
            &query.str_match.pandas(dataframe, 'natoms', natoms)
            &query.str_match.pandas(dataframe, 'natypes', natypes)
        )
        return matches

    @staticmethod
    def mongoquery(name=None, key=None,
                   method='dynamic', standing='good',
                   family=None, parent_key=None, 
                   potential_LAMMPS_id=None, potential_LAMMPS_key=None,
                   potential_id=None, potential_key=None,
                   crystalfamily=None, composition=None,
                   symbols=None, natoms=None, natypes=None):

        mquery = {}
        query.str_match.mongo(mquery, f'name', name)
        root = f'content.{modelroot}'

        query.str_match.mongo(mquery, f'{root}.key', key)
        query.str_match.mongo(mquery, f'{root}.method', method)
        query.str_match.mongo(mquery, f'{root}.standing', standing)
        query.str_match.mongo(mquery, f'{root}.system-info.family', family)
        query.str_match.mongo(mquery, f'{root}.system-info.parent_key', parent_key)
        query.str_match.mongo(mquery, f'{root}.potential-LAMMPS.id', potential_LAMMPS_id)
        query.str_match.mongo(mquery, f'{root}.potential-LAMMPS.key', potential_LAMMPS_key)
        query.str_match.mongo(mquery, f'{root}.potential-LAMMPS.potential.id', potential_id)
        query.str_match.mongo(mquery, f'{root}.potential-LAMMPS.potential.key', potential_key)
        query.str_match.mongo(mquery, f'{root}.system-info.cell.crystal-family', crystalfamily)
        query.str_match.mongo(mquery, f'{root}.system-info.cell.composition', composition)
        query.in_list.mongo(mquery, f'{root}.system-info.symbol', symbols)
        query.str_match.mongo(mquery, f'{root}.atomic-system.atoms.natoms', natoms)
        query.str_match.mongo(mquery, f'{root}.system-info.cell.natypes', natypes)

        return mquery

    @staticmethod
    def cdcsquery(key=None, method='dynamic', standing='good',
                  family=None, parent_key=None, 
                  potential_LAMMPS_id=None, potential_LAMMPS_key=None,
                  potential_id=None, potential_key=None,
                  crystalfamily=None, composition=None, 
                  symbols=None, natoms=None, natypes=None):

        mquery = {}
        root = modelroot
        
        query.str_match.mongo(mquery, f'{root}.key', key)
        query.str_match.mongo(mquery, f'{root}.method', method)
        query.str_match.mongo(mquery, f'{root}.standing', standing)
        query.str_match.mongo(mquery, f'{root}.system-info.family', family)
        query.str_match.mongo(mquery, f'{root}.system-info.parent_key', parent_key)
        query.str_match.mongo(mquery, f'{root}.potential-LAMMPS.id', potential_LAMMPS_id)
        query.str_match.mongo(mquery, f'{root}.potential-LAMMPS.key', potential_LAMMPS_key)
        query.str_match.mongo(mquery, f'{root}.potential-LAMMPS.potential.id', potential_id)
        query.str_match.mongo(mquery, f'{root}.potential-LAMMPS.potential.key', potential_key)
        query.str_match.mongo(mquery, f'{root}.system-info.cell.crystal-family', crystalfamily)
        query.str_match.mongo(mquery, f'{root}.system-info.cell.composition', composition)
        query.in_list.mongo(mquery, f'{root}.system-info.symbol', symbols)
        query.str_match.mongo(mquery, f'{root}.atomic-system.atoms.natoms', natoms)
        query.str_match.mongo(mquery, f'{root}.system-info.cell.natypes', natypes)

        return mquery


# coding: utf-8
        
def get_relaxed_crystals(self, local=None, remote=None, name=None, key=None,
                         method='dynamic', standing='good',
                         family=None, parent_key=None, 
                         potential_LAMMPS_id=None, potential_LAMMPS_key=None,
                         potential_id=None, potential_key=None,
                         symbols=None, natoms=None, natypes=None,
                         refresh_cache=False, return_df=False, verbose=False):
    """
    Get all matching relaxed crystals from the database.
    
    Parameters
    ----------
    local : bool, optional
        Indicates if the local location is to be searched.  Default value
        matches the value set when the database was initialized.
    remote : bool, optional
        Indicates if the remote location is to be searched.  Default value
        matches the value set when the database was initialized.
    name : str or list
        The record name(s) to parse by.  For relaxed crystal records, the
        names should correspond to the key.
    key : str or list, optional
        UUID4 key(s) to search for.  Each entry has a unique random-generated
        UUID4 key.
    method : str or list or None, optional
        The relaxation method used.  Allowed values are dynamic, static and box.
        Default value is dynamic (the most rigorous relaxation method).  All
        will be loaded if set to None.
    standing : str or list or None, optional
        "good" records are the unique crystals found with the most rigorous
        relaxation, and with known prototypes over DFT structures.  "bad" are
        records filtered out, usually for being duplicates.  Default value is
        "good".  All will be loaded if set to None.
    family : str or atomman.library.CrystalPrototype or list, optional
        The crystal family associated with the relaxed crystal - either crystal
        prototype name or MP/OQMD database entry name.
    parent_key : str or list, optional
        The UUID4 key(s) assigned to the calculation that the record is based
        on.
    potential_LAMMPS_id : str or list, optional
        The id for a LAMMPS implemented potential to limit the search by.
    potential_LAMMPS_key : str or list, optional
        The UUID4 for a LAMMPS implemented potential to limit the search by.
    potential_id : str or list, optional
        The id for a potential to limit the search by.
    potential_key : str or list, optional
        The UUID4 for a potential to limit the search by.
    symbols : str or list, optional
        Element symbols in the crystal to limit the search by.
    natypes : int or list, optional
        The number(s) of unique atom types to limit the search by.
    natoms : int or list, optional
        The number of unique atoms in the crystal's unit cell to limit the
        search by.
    refresh_cache : bool, optional
        If the local database is of style "local", indicates if the metadata
        cache file is to be refreshed.  If False,
        metadata for new records will be added but the old record metadata
        fields will not be updated.  If True, then the metadata for all
        records will be regenerated, which is needed to update the metadata
        for modified records.
    return_df : bool, optional
        If True, then the corresponding pandas.Dataframe of metadata
        will also be returned.
    verbose : bool, optional
        If True, info messages will be printed during operations.  Default
        value is False.
    """

    return self.get_records('relaxed_crystal', local=local, remote=remote, name=name,
                            key=key, method=method, standing=standing, family=family,
                            parent_key=parent_key, potential_LAMMPS_id=potential_LAMMPS_id,
                            potential_LAMMPS_key=potential_LAMMPS_key,
                            potential_id=potential_id, potential_key=potential_key,
                            symbols=symbols, natoms=natoms, natypes=natypes,
                            refresh_cache=refresh_cache, return_df=return_df, verbose=verbose)

def get_relaxed_crystal(self, local=None, remote=None, name=None,
                        key=None, method='dynamic', standing='good',
                        family=None, parent_key=None, 
                        potential_LAMMPS_id=None, potential_LAMMPS_key=None,
                        potential_id=None, potential_key=None,
                        symbols=None, natoms=None, natypes=None, keyword=None,
                        prompt=True, refresh_cache=False, verbose=False):
    """
    Retrieves exactly one matching relaxed crystal from the database.
    
    Parameters
    ----------
    local : bool, optional
        Indicates if the local location is to be searched.  Default value
        matches the value set when the database was initialized.
    remote : bool, optional
        Indicates if the remote location is to be searched.  Default value
        matches the value set when the database was initialized.
    name : str or list
        The record name(s) to parse by.  For relaxed crystal records, the
        names should correspond to the key.
    key : str or list, optional
        UUID4 key(s) to search for.  Each entry has a unique random-generated
        UUID4 key.
    method : str or list or None, optional
        The relaxation method used.  Allowed values are dynamic, static and box.
        Default value is dynamic (the most rigorous relaxation method).  All
        will be loaded if set to None.
    standing : str or list or None, optional
        "good" records are the unique crystals found with the most rigorous
        relaxation, and with known prototypes over DFT structures.  "bad" are
        records filtered out, usually for being duplicates.  Default value is
        "good".  All will be loaded if set to None.
    family : str or atomman.library.CrystalPrototype or list, optional
        The crystal family associated with the relaxed crystal - either crystal
        prototype name or MP/OQMD database entry name.
    parent_key : str or list, optional
        The UUID4 key(s) assigned to the calculation that the record is based
        on.
    potential_LAMMPS_id : str or list, optional
        The id for a LAMMPS implemented potential to limit the search by.
    potential_LAMMPS_key : str or list, optional
        The UUID4 for a LAMMPS implemented potential to limit the search by.
    potential_id : str or list, optional
        The id for a potential to limit the search by.
    potential_key : str or list, optional
        The UUID4 for a potential to limit the search by.
    symbols : str or list, optional
        Element symbols in the crystal to limit the search by.
    natypes : int or list, optional
        The number of unique element model symbols in the crystal to limit
        the search by.
    natypes : int or list, optional
        The number(s) of unique atom types to limit the search by.
    prompt : bool
        If prompt=True (default) then a screen input will ask for a selection
        if multiple matching potentials are found.  If prompt=False, then an
        error will be thrown if multiple matches are found.
    refresh_cache : bool, optional
        If the local database is of style "local", indicates if the metadata
        cache file is to be refreshed.  If False,
        metadata for new records will be added but the old record metadata
        fields will not be updated.  If True, then the metadata for all
        records will be regenerated, which is needed to update the metadata
        for modified records.
    verbose : bool, optional
        If True, info messages will be printed during operations.  Default
        value is False.
    """
    def promptfxn(df):
        header = '#  family               symbols  alat    Ecoh    method  standing'
        print(header)

        js = df.sort_values('cohesive_energy').index
        for i, j in enumerate(js):
            crystal = df.loc[j]
            row =  f'{i+1:2} {crystal.family:20.20} '
            row += f'{"".join(crystal.symbols):8.8} '
            row += f'{crystal.a:7.4f} '
            row += f'{crystal.cohesive_energy:7.4f} '
            row += f'{crystal.method:7.7} '
            row += f'{crystal.standing:4.4}'
            print(row)
        
        i = int(input('Please select one:')) - 1
        if i < 0 or i >= len(js):
            raise ValueError('Invalid selection')

        return js[i]

    return self.get_record('relaxed_crystal', local=local, remote=remote, name=name,
                           key=key, method=method, standing=standing, family=family,
                           parent_key=parent_key, potential_LAMMPS_id=potential_LAMMPS_id,
                           potential_LAMMPS_key=potential_LAMMPS_key,
                           potential_id=potential_id, potential_key=potential_key,
                           symbols=symbols, natoms=natoms, natypes=natypes,
                           prompt=prompt, promptfxn=promptfxn, refresh_cache=refresh_cache, 
                           verbose=verbose)

def download_relaxed_crystals(self, name=None, key=None, method='dynamic',
                              standing='good', family=None, parent_key=None, 
                              potential_LAMMPS_id=None, potential_LAMMPS_key=None,
                              potential_id=None, potential_key=None,
                              symbols=None, natoms=None, natypes=None, keyword=None,
                              overwrite=False, verbose=False):
    """
    Download citation records from the remote and save to localpath.
    
    Parameters
    ----------
    name : str or list
        The record name(s) to parse by.  For relaxed crystal records, the
        names should correspond to the key.
    key : str or list, optional
        UUID4 key(s) to search for.  Each entry has a unique random-generated
        UUID4 key.
    method : str or list or None, optional
        The relaxation method used.  Allowed values are dynamic, static and box.
        Default value is dynamic (the most rigorous relaxation method).  All
        will be loaded if set to None.
    standing : str or list or None, optional
        "good" records are the unique crystals found with the most rigorous
        relaxation, and with known prototypes over DFT structures.  "bad" are
        records filtered out, usually for being duplicates.  Default value is
        "good".  All will be loaded if set to None.
    family : str or atomman.library.CrystalPrototype or list, optional
        The crystal family associated with the relaxed crystal - either crystal
        prototype name or MP/OQMD database entry name.
    parent_key : str or list, optional
        The UUID4 key(s) assigned to the calculation that the record is based
        on.
    potential_LAMMPS_id : str or list, optional
        The id for a LAMMPS implemented potential to limit the search by.
    potential_LAMMPS_key : str or list, optional
        The UUID4 for a LAMMPS implemented potential to limit the search by.
    potential_id : str or list, optional
        The id for a potential to limit the search by.
    potential_key : str or list, optional
        The UUID4 for a potential to limit the search by.
    symbols : str or list, optional
        Element symbols in the crystal to limit the search by.
    natypes : int or list, optional
        The number of unique element model symbols in the crystal to limit
        the search by.
    natoms : int or list, optional
        The number of unique atoms in the crystal's unit cell to limit the
        search by.
    overwrite : bool, optional
        Flag indicating if any existing local records with names matching
        remote records are updated (True) or left unchanged (False).  Default
        value is False.
    verbose : bool, optional
        If True, info messages will be printed during operations.  Default
        value is False.
    """

    self.download_records('relaxed_crystal', name=name,
                          key=key, method=method, standing=standing, family=family,
                          parent_key=parent_key, potential_LAMMPS_id=potential_LAMMPS_id,
                          potential_LAMMPS_key=potential_LAMMPS_key,
                          potential_id=potential_id, potential_key=potential_key,
                          symbols=symbols, natoms=natoms, natypes=natypes,
                          overwrite=overwrite, verbose=verbose)


from copy import deepcopy

# https://github.com/usnistgov/DataModelDict
from DataModelDict import DataModelDict as DM

from datamodelbase.record import Record
from datamodelbase import query

from ... import System

modelroot = 'crystal-prototype'

class CrystalPrototype(Record):
    
    @property
    def style(self):
        """str: The record style"""
        return 'crystal_prototype'

    @property
    def modelroot(self):
        """str: The root element of the content"""
        return modelroot

    @property
    def xsd_filename(self):
        return ('atomman.library.xsd', f'{self.style}.xsd')

    def load_model(self, model, name=None):
        super().load_model(model, name=name)        
        proto = self.model[modelroot]
        
        self.__key = proto['key']
        self.__id = proto['id']
        self.__commonname = proto['name']
        self.__prototype = proto['prototype']
        self.__pearson = proto['Pearson-symbol']
        self.__strukturbericht = proto['Strukturbericht']
        
        self.__sg_number = proto['space-group']['number']
        self.__sg_hm = proto['space-group']['Hermann-Maguin']
        self.__sg_schoenflies = proto['space-group']['Schoenflies']
        
        self.__crystalfamily = proto['system-info']['cell']['crystal-family']
        self.__natypes = proto['system-info']['cell']['natypes']
        self.__ucell = None

        # Set name as id if no name given
        try:
            self.name
        except:
            self.name = self.id

    @property
    def id(self):
        return self.__id

    @property
    def key(self):
        return self.__key

    @property
    def commonname(self):
        return self.__commonname
    
    @property
    def prototype(self):
        return self.__prototype
    
    @property
    def pearson(self):
        return self.__pearson

    @property
    def strukturbericht(self):
        return self.__strukturbericht

    @property
    def sg_number(self):
        return self.__sg_number

    @property
    def sg_hm(self):
        return self.__sg_hm

    @property
    def sg_schoenflies(self):
        return self.__sg_schoenflies

    @property
    def crystalfamily(self):
        return self.__crystalfamily

    @property
    def natypes(self):
        return self.__natypes

    @property
    def ucell(self):
        if self.__ucell is None:
            self.__ucell = System(model=self.model)
        return self.__ucell

    def build_model(self):
        return self.model

    def metadata(self):
        """Returns the simple metadata fields of the record as a flat dict."""
        params = {}
        params['name'] = self.name
        params['key'] = self.key
        params['id'] = self.id
        params['commonname'] = self.commonname
        params['prototype'] = self.prototype
        params['pearson'] = self.pearson
        params['strukturbericht'] = self.strukturbericht
        params['sg_number'] = self.sg_number
        params['sg_hm'] = self.sg_hm
        params['sg_schoenflies'] = self.sg_schoenflies
        params['crystalfamily'] = self.crystalfamily
        params['natypes'] = self.natypes
        
        return params

    @staticmethod
    def pandasfilter(dataframe, name=None, id=None,
                     key=None, commonname=None, prototype=None,
                     pearson=None, strukturbericht=None, sg_number=None,
                     sg_hm=None, sg_schoenflies=None, crystalfamily=None,
                     natypes=None):
        matches = (
            query.str_match.pandas(dataframe, 'name', name)
            &query.str_match.pandas(dataframe, 'key', key)
            &query.str_match.pandas(dataframe, 'id', id)
            &query.str_match.pandas(dataframe, 'commonname', commonname)
            &query.str_match.pandas(dataframe, 'prototype', prototype)
            &query.str_match.pandas(dataframe, 'pearson', pearson)
            &query.str_match.pandas(dataframe, 'strukturbericht', strukturbericht)
            &query.str_match.pandas(dataframe, 'sg_number', sg_number)
            &query.str_match.pandas(dataframe, 'sg_hm', sg_hm)
            &query.str_match.pandas(dataframe, 'sg_schoenflies', sg_schoenflies)
            &query.str_match.pandas(dataframe, 'crystalfamily', crystalfamily)
            &query.str_match.pandas(dataframe, 'natypes', natypes)
        )
        return matches

    @staticmethod
    def mongoquery(name=None, id=None, key=None, commonname=None,
                   prototype=None, pearson=None, strukturbericht=None,
                   sg_number=None, sg_hm=None, sg_schoenflies=None,
                   crystalfamily=None, natypes=None):

        mquery = {}
        query.str_match.mongo(mquery, f'name', name)
        root = f'content.{modelroot}'

        query.str_match.mongo(mquery, f'{root}.key', key)
        query.str_match.mongo(mquery, f'{root}.id', id)
        query.str_match.mongo(mquery, f'{root}.name', commonname)
        query.str_match.mongo(mquery, f'{root}.prototype', prototype)
        query.str_match.mongo(mquery, f'{root}.Pearson-symbol', pearson)
        query.str_match.mongo(mquery, f'{root}.Strukturbericht', strukturbericht)
        query.str_match.mongo(mquery, f'{root}.space-group.number', sg_number)
        query.str_match.mongo(mquery, f'{root}.space-group.Hermann-Maguin', sg_hm)
        query.str_match.mongo(mquery, f'{root}.space-group.Schoenflies', sg_schoenflies)
        query.str_match.mongo(mquery, f'{root}.system-info.cell.crystal-family', crystalfamily)
        query.str_match.mongo(mquery, f'{root}.system-info.cell.natypes', natypes)

        return mquery

    @staticmethod
    def cdcsquery(id=None, key=None, commonname=None,
                  prototype=None, pearson=None, strukturbericht=None,
                  sg_number=None, sg_hm=None, sg_schoenflies=None,
                  crystalfamily=None, natypes=None):

        mquery = {}
        root = modelroot

        query.str_match.mongo(mquery, f'{root}.key', key)
        query.str_match.mongo(mquery, f'{root}.id', id)
        query.str_match.mongo(mquery, f'{root}.name', commonname)
        query.str_match.mongo(mquery, f'{root}.prototype', prototype)
        query.str_match.mongo(mquery, f'{root}.Pearson-symbol', pearson)
        query.str_match.mongo(mquery, f'{root}.Strukturbericht', strukturbericht)
        query.str_match.mongo(mquery, f'{root}.space-group.number', sg_number)
        query.str_match.mongo(mquery, f'{root}.space-group.Hermann-Maguin', sg_hm)
        query.str_match.mongo(mquery, f'{root}.space-group.Schoenflies', sg_schoenflies)
        query.str_match.mongo(mquery, f'{root}.system-info.cell.crystal-family', crystalfamily)
        query.str_match.mongo(mquery, f'{root}.system-info.cell.natypes', natypes)
        
        return mquery


# coding: utf-8
# Standard Python libraries
import datetime

from DataModelDict import DataModelDict as DM

# https://numpy.org/
import numpy as np

# https://pandas.pydata.org/
import pandas as pd

import matplotlib.pyplot as plt

# atomman imports 
from .. import Atoms, Box, System
from ..tools import aslist
import atomman.unitconvert as uc

class BondAngleMap():
    """
    Class for generating and analyzing energies of three atom clusters that
    explore a range of interatomic distances and bond angles.  Can be used to
    characterize the 3 atom bond nature of interatomic potentials.
    """
    
    def __init__(self, model=None, rmin=None, rmax=None, rnum=None,
                 thetamin=None, thetamax=None, thetanum=None,
                 r_ij=None, r_ik=None, theta=None, energy=None, symbols=None):
        """
        Class initializer.  The cluster coordinates (r distances and theta
        angles) are required and can be specified in one of three ways.
        - A data model containing all the information.
        - Input ranges (rmin, rmax, rnum, thetamin, thetamax, thetanum).  The r
          range parameters will be used to generate values for both r_ij and
          r_ik.
        - Explicitly giving r_ij, r_ik and theta values.

        Parameters
        ----------
        model : str, path-like object or DataModelDict, optional
            Collected data model results from a series of runs.  Contains both
            coordinate information and energy values.
        rmin : float, optional
            The minimum value used for the r_ij and r_ik spacings.
        rmax : float, optional
            The maximum value used for the r_ij and r_ik spacings.
        rnum : float, optional
            The number of values used for the r_ij and r_ik spacings.
        thetamin : float, optional
            The minimum value used for the theta angles.
        thetamax : float, optional
            The maximum value used for the theta angles.
        thetanum : float, optional
            The number of values used for the theta angles.
        r_ij : array-like, optional
            All r_ij values used.  If given, the lengths of r_ij, r_ik and
            theta need to be the same.
        r_ik : array-like, optional
            All r_ik values used.  If given, the lengths of r_ij, r_ik and
            theta need to be the same.
        theta : array-like, optional
            All theta values used.  If given, the lengths of r_ij, r_ik and
            theta need to be the same.
        energy : array-like, optional
            All measured energies.  If r_ij, r_ik and theta are given then
            all should be the same length.  If the coordinate range parameters
            are given, then the energies should be of length rnum*rnum*thetanum
            and ordered to correspond to three embedded loops with r_ij
            iterating in the outside loop, r_ik in the middle and theta in the
            inside.  If energy is not given, then all values will initially be
            set to np.nan.
        symbols : str or list, optional
            Element model symbol(s) to associate with the three atoms if/when
            systems are created.  Can either be a single symbol to assign to 
            all atoms, or three symbols to assign to atoms i, j, and k
            individually.  Not needed if systems are not generated by this
            class.
        """
        self.symbols = symbols
        if model is not None:
            try:
                assert rmin is None and rmax is None and rnum is None
                assert thetamin is None and thetamax is None and thetanum is None
                assert r_ij is None and r_ik is None and theta is None
                assert energy is None
            except:
                raise ValueError('model cannot be given energy or coordinate parameters')
            self.model(model)
        else:
            self.set(energy=energy, rmin=rmin, rmax=rmax, rnum=rnum,
                     thetamin=thetamin, thetamax=thetamax, thetanum=thetanum,
                     r_ij=r_ij, r_ik=r_ik, theta=theta)

    @property
    def rmin(self):
        """float or None: The minimum value used for the r_ij and r_ik spacings."""
        return self.__rmin

    @property
    def rmax(self):
        """float or None: The maximum value used for the r_ij and r_ik spacings."""
        return self.__rmax

    @property
    def rnum(self):
        """int or None: The number of values used for the r_ij and r_ik spacings."""
        return self.__rnum

    @property
    def thetamin(self):
        """float or None: The minimum value used for the theta angles."""
        return self.__thetamin

    @property
    def thetamax(self):
        """float or None: The maximum value used for the theta angles."""
        return self.__thetamax

    @property
    def thetanum(self):
        """int or None: The number of values used for the theta angles."""
        return self.__thetanum

    @property
    def df(self):
        """pandas.Dataframe : The cluster coordinates and energies."""
        return self.__df

    @property
    def symbols(self):
        """list or None: the atomic symbols associated with the three atoms"""
        return self.__symbols
    
    @symbols.setter
    def symbols(self, value):
        
        if value is not None:
            value = aslist(value)
            if len(value) != 1 and len(value) != 3:
                raise ValueError('Number of symbols must be 1 or 3')
        
        self.__symbols = value

    def model(self, model=None, length_unit='angstrom', energy_unit='eV'):
        """
        Loads or generates a bond angle map data model.

        Note: Generating data models is currently limited to regular values,
        i.e. ones which the coordinates correspond to embedded loops with
        r_ij iterating in the outside loop, r_ik in the middle loop and 
        theta in the inside loop.

        Parameters
        ----------
        model : str, file-like object or DataModelDict, optional
            The data model content or file containing the bond angle map data.
            If given, the content will be read in and set to the current object.
            If not given, then a data model will be returned for the object.
        length_unit : str, optional
            The unit of length to save the rmin and max values in when
            generating a data model.  Default is 'angstrom'.
            value
        energy_unit : str, optional
            The unit of energy to save the energy values in when generating a
            data model.  Default value is 'eV'.
        
        Returns
        -------
        DataModelDict.DataModelDict
            The data model containing the bond angle map coordinate information
            and measured energies.  Only returned if model is not given as a
            parameter.

        Raises
        ------
        ValueError
            If the data is irregular, i.e. coordinates do not conform to
            embedded loops with r_ij in the outer loop, r_ik in the middle loop
            and theta in the inside loop.
        """
        # Read in a data model
        if model is not None:
            cluster = DM(model).find('bond-angle-map')
            
            rmin = uc.value_unit(cluster['minimum-r-spacing'])
            rmax = uc.value_unit(cluster['maximum-r-spacing'])
            rnum = int(cluster['number-of-r-spacings'])

            thetamin = float(cluster['minimum-angle'])
            thetamax = float(cluster['maximum-angle'])
            thetanum = int(cluster['number-of-angles'])

            energy = uc.value_unit(cluster['energy'])

            self.set(energy=energy, rmin=rmin, rmax=rmax, rnum=rnum, 
                     thetamin=thetamin, thetamax=thetamax, thetanum=thetanum)
        
        else:
            if self.rnum is None:
                raise ValueError('model does not support irregular measurements')
            
            model = DM()
            model['bond-angle-map'] = cluster = DM()
            
            cluster['minimum-r-spacing'] = uc.model(self.rmin, length_unit)
            cluster['maximum-r-spacing'] = uc.model(self.rmax, length_unit)
            cluster['number-of-r-spacings'] = self.rnum

            cluster['minimum-angle'] = self.thetamin
            cluster['maximum-angle'] = self.thetamax
            cluster['number-of-angles'] = self.thetanum

            cluster['energy'] = uc.model(self.df.energy, energy_unit)

            return model

    def set(self, rmin=None, rmax=None, rnum=None,
            thetamin=None, thetamax=None, thetanum=None,
            r_ij=None, r_ik=None, theta=None, energy=None):
        """
        Sets the bond angle coordinates and the associated energies, if given.

        Parameters
        ----------
        energy : array-like, optional
            All measured energies.  If r_ij, r_ik and theta are given then
            all should be the same length.  If the range parameters are given
            then the length should be thetanum*rnum*rnum.  If not given, then
            a new array of nan values will be constructed.
        rmin : float, optional
            The minimum value used for the r_ij and r_ik spacings.
        rmax : float, optional
            The maximum value used for the r_ij and r_ik spacings.
        rnum : float, optional
            The number of values used for the r_ij and r_ik spacings.
        thetamin : float, optional
            The minimum value used for the theta angles.
        thetamax : float, optional
            The maximum value used for the theta angles.
        thetanum : float, optional
            The number of values used for the theta angles.
        r_ij : array-like, optional
            All r_ij values used.  If given, the lengths of r_ij, r_ik and
            theta need to be the same.
        r_ik : array-like, optional
            All r_ik values used.  If given, the lengths of r_ij, r_ik and
            theta need to be the same.
        theta : array-like, optional
            All theta values used.  If given, the lengths of r_ij, r_ik and
            theta need to be the same.
        energy : array-like, optional
            All measured energies.  If r_ij, r_ik and theta are given then
            all should be the same length.  If the coordinate range parameters
            are given, then the energies should be of length rnum*rnum*thetanum
            and ordered to correspond to three embedded loops with r_ij
            iterating in the outside loop, r_ik in the middle and theta in the
            inside.  If energy is not given, then all values will initially be
            set to np.nan.
        """

        # Set coordinate values based on ranges
        if rmin is not None:
            if r_ij is not None or r_ik is not None or theta is not None:
                raise ValueError('range parameters and explicit values cannot be mixed')
            try:
                rvals = np.linspace(rmin, rmax, rnum)
                tvals = np.linspace(thetamin, thetamax, thetanum)
            except:
                raise ValueError('Invalid range parameters')

            # Set range parameters as class properties
            self.__rmin = rmin
            self.__rmax = rmax
            self.__rnum = rnum
            self.__thetamin = thetamin
            self.__thetamax = thetamax
            self.__thetanum = thetanum

            # Generate the input parameters
            r_ij, r_ik, theta = np.meshgrid(rvals, rvals, tvals, indexing='ij')
            r_ij = r_ij.flatten()
            r_ik = r_ik.flatten()
            theta = theta.flatten()

        # Explicitly set coordinate values
        else:
            try:
                assert rmax is None and rnum is  None
                assert thetamin is None and thetamax is None and thetanum is None
            except:
                raise ValueError('range parameters and explicit values cannot be mixed') 
            try:
                if len(r_ij) != len(r_ik) or len(r_ij) != len(theta):
                    raise ValueError('Equal numbers of r_ij, r_ik, and theta values must be given')
            except:
                raise ValueError('Invalid parameters')

            r_ij = np.asarray(r_ij)
            r_ik = np.asarray(r_ik)
            theta = np.asarray(theta)
            
            # Try to extract range parameters
            rmin = r_ij.min()
            rmax = r_ij.max()
            rnum = len(np.unique(r_ij))

            thetamin = theta.min()
            thetamax = theta.max()
            thetanum = len(np.unique(theta))

            try:
                assert len(r_ij) == thetanum * rnum * rnum
                rvals = np.linspace(rmin, rmax, rnum)
                tvals = np.linspace(thetamin, thetamax, thetanum)
                r_ij_test, r_ik_test, theta_test = np.meshgrid(rvals, rvals, tvals, indexing='ij')
                assert np.allclose(r_ij, r_ij_test.flatten())
                assert np.allclose(r_ik, r_ik_test.flatten())
                assert np.allclose(theta, theta_test.flatten())
            except:
                # Set range parameters as None to indicate not correctly ordered
                self.__rmin = None
                self.__rmax = None
                self.__rnum = None
                self.__thetamin = None
                self.__thetamax = None
                self.__thetanum = None
            else:
                # Set range parameters
                self.__rmin = rmin
                self.__rmax = rmax
                self.__rnum = rnum
                self.__thetamin = thetamin
                self.__thetamax = thetamax
                self.__thetanum = thetanum
        
        # Check energy values
        if energy is None:
            energy = np.full(len(r_ij), np.nan)
        elif len(energy) != len(r_ij):
            raise ValueError('Mismatch between number of energies given and expected')

        # Compute r_jk values
        r_jk = np.sqrt(r_ij**2 + r_ik**2 - r_ij * r_ik * 2 * np.cos(np.radians(theta)))

        # Build DataFrame
        df = {}
        df['r_ij'] = r_ij
        df['r_ik'] = r_ik
        df['r_jk'] = r_jk
        df['theta'] = theta
        df['energy'] = energy
        self.__df = pd.DataFrame(df)

    def itercoords(self):
        """
        Iterates through the three-body coordinates, which can be used as inputs for
        computing energies.

        Yields
        ------
        r_ij : float
            The radial distance between atoms i and j.
        r_ik : float
            The radial distance between atoms i and k.
        r_jk : float
            The radial distance between atoms j and k.
        theta : float
            The angle between i-j and i-k in degrees.
        """
        for i in self.df.index:
            series = self.df.loc[i]
            yield series.r_ij, series.r_ik, series.r_jk, series.theta

    def itersystem(self, symbols=None, copy=False):
        """
        Iterates through the three-body coordinates and returns a System for each.
        Useful for generating configuration files for simulators.  The atom
        coordinates will be set such that atom 0 is at [0,0,0], atom 1 at
        [r_ij,0,0] and atom 2 is in the xy plane based on r_ik and theta.

        Parameters
        ----------
        symbols : str or list, optional
            The element model symbols to assign to the atoms.  Can either be
            one value for all atoms, or three values for each atom individually.
            If not given here, will use the values set during class
            initialization.
        copy : bool, optional
            If False (default), then the yielded system is the same object with
            the coordinates shifted.  If True, each yielded system is a new object.

        Yields
        ------
        atomman.System
            The atomic system containing the three-body cluster. 
        """
        # Set symbols
        if symbols is not None:
            self.symbols = symbols 
        symbols = self.symbols

        # Identify box bounds based on r_ij values
        rhi = 3 * self.df.r_ij.max()
        rlo = - rhi
        
        # Identify atypes from symbols
        if symbols is None or len(symbols) == 1:
            atype = np.array([1,1,1])
        elif len(symbols) == 3:
            symbols, atype = np.unique(symbols, return_inverse=True)
        else:
            print(symbols)
            raise ValueError('Invalid symbols somehow...')

        # Copy = True means generate new system each iteration
        if copy:
            for r_ij, r_ik, r_jk, theta in self.itercoords():

                # Build the pos array
                j_x = r_ij
                k_x = r_ik * np.cos(np.radians(theta))
                k_y = r_ik * np.sin(np.radians(theta))
                pos = np.array([[0.0, 0.0, 0.0],
                                [j_x, 0.0, 0.0],
                                [k_x, k_y, 0.0]])

                # Build and yield a new system
                box = Box(xlo=rlo, xhi=rhi, ylo=rlo, yhi=rhi, zlo=-1.0, zhi=1.0)
                atoms = Atoms(atype=atype, pos=pos)
                yield System(atoms=atoms, box=box, symbols=symbols, pbc=[False, False, False])

        # Copy = False means only generate one system and modify pos
        else:
            # Build system with all coordinates at [0,0,0]
            box = Box(xlo=rlo, xhi=rhi, ylo=rlo, yhi=rhi, zlo=-1.0, zhi=1.0)
            atoms = Atoms(atype=atype, pos=np.zeros([3,3]))
            system = System(atoms=atoms, box=box, symbols=symbols, pbc=[False, False, False])
        
            for r_ij, r_ik, r_jk, theta in self.itercoords():
                
                # Modify the three non-zero coordinates
                system.atoms.pos[1,0] = r_ij
                system.atoms.pos[2,0] = r_ik * np.cos(np.radians(theta))
                system.atoms.pos[2,1] = r_ik * np.sin(np.radians(theta))

                yield system
            
    def save_table(self, filename, include_header=True):
        """
        Saves a tabulated representation of the coordinates and energy values to a file.

        Parameters
        ----------
        filename : str
            The path to the file where the table will be saved.
        include_header : bool
            If True (default) then header comments will be listed at the top of the file.
        """
        with open(filename, 'w') as f:

            if include_header:
                emin = self.df.energy.min()

                # Create the header comment lines
                f.write(f'# Comment: this is a file containing an r12, r13, theta - energy-block E(r12, r13, theta) E_min={emin:18.14} eV\n')
                f.write(f'# File format: After 9 comment lines (starting with #) there are three lines with (angle in degrees):\n')
                f.write(f'# r12_min, r12_max, n_bins_r12; r13_min, r13_max, n_bins_r13, theta_min, theta_max, n_bins_theta followed by\n')
                f.write(f'# the data block in the format index_r12, index_r13_ index_theta, E(i_r12,i_r13,i_theta) [eV] one per line.\n')   
                f.write(f'# Conversion from index to value: val = val_min + (index-1)*(val_max - val_min)/(n_bins-1)\n')
                f.write(f'# file created by xxx via code atomman on date {datetime.date.today}\n')
                #f.write(f'# This is an example for 3-body MD-potential C-W; Reference: J. Jones, CPPC 34, p.123-432 (2019)\n')
                #f.write(f'# Atom 1: C, Atom 2: W, Atom 3: W\n')
                #f.write(f'# Version 0.2 (2021-02-15) by U v. Toussaint (IPP)\n')
                f.write(f'{self.rmin:18.14} {self.rmax:18.14} {self.rnum}\n')
                f.write(f'{self.rmin:18.14} {self.rmax:18.14} {self.rnum}\n')
                f.write(f'{self.thetamin:18.14} {self.thetamax:18.14} {self.thetanum}\n')

            l = 0
            for i in range(self.rnum):
                for j in range(self.rnum):
                    for k in range(self.thetanum):
                        f.write(f'{i+1:6} {j+1:6} {k+1:6} {self.df.energy.values[l]:18.14}\n')
                        l += 1

    def pdf(self, nbins=301, energymin=-15.0, energymax=15.0):
        """
        Returns the probability density function for the energy
        
        Parameters
        ----------
        nbins : int, optional
            The number of histogram bins to use.  Default value is 301.
        energymin : float, optional
            The minimum energy bound to consider.  Default value is -15.0.
        energymax : float, optional
            The maximum energy bound to consider.  Default value is 15.0.

        Returns
        -------
        pdf : numpy.NDArray
            The probability density function associated with each bin.
        centers : numpy.NDArray
            The center values for each bin.
        """

        hist, edges = np.histogram(self.df.energy, bins=nbins, range=(energymin, energymax))

        # Divide the historgram count by total number of measurements
        pdf = hist/len(self.df)

        # Average the bin edges to get the bin centers
        centers = (edges[:-1] + edges[1:]) / 2

        return pdf, centers

    def cumulative_pdf(self, nbins=301, energymin=-15.0, energymax=15.0):
        """
        Returns the cumulative probability density function for the energy.
        
        Parameters
        ----------
        nbins : int, optional
            The number of histogram bins to use.  Default value is 301.
        energymin : float, optional
            The minimum energy bound to consider.  Default value is -15.0.
        energymax : float, optional
            The maximum energy bound to consider.  Default value is 15.0.

        Returns
        -------
        cum_pdf : numpy.NDArray
            The cumulative probability density function associated with each bin.
        centers : numpy.NDArray
            The center values for each bin.
        """

        # Get pdf and centers in the range
        pdf, centers = self.pdf(nbins=nbins, energymin=energymin, energymax=energymax)

        # Calculate cumulative pdf below energymin
        shift = np.sum(self.df.energy < energymin) / len(self.df)

        # Calculate the cumulative sum of pdf and apply the shift
        cum_pdf = np.cumsum(pdf) + shift

        return cum_pdf, centers

    def plot_pdf(self, nbins=301, energymin=-15.0, energymax=15.0,
                matplotlib_axes=None, **kwargs):
        """
        Generates a plot of the probability density function of the energy.

        Parameters
        ----------
        nbins : int, optional
            The number of histogram bins to use.  Default value is 301.
        energymin : float, optional
            The minimum energy bound to consider.  Default value is -15.0.
        energymax : float, optional
            The maximum energy bound to consider.  Default value is 15.0.
        matplotlib_axes : matplotlib.Axes.axes, optional, optional
            An existing plotting axis to add the pdf plot to.  If not given,
            a new figure object will be generated.
        **kwargs : any, optional
            Any additional key word arguments will be passed to
            matplotlib.pyplot.figure for generating a new figure object (if
            axis is not given).
        
        Returns
        -------
        matplotlib.Figure
            The generated figure.  Not returned if matplotlib_axes is given.
        """
        # Get pdf and centers in the range
        pdf, centers = self.pdf(nbins=nbins, energymin=energymin, energymax=energymax)

        # Initial plot setup and parameters
        if matplotlib_axes is None:
            fig = plt.figure(**kwargs)
            ax1 = fig.add_subplot(111)
        else:
            ax1 = matplotlib_axes
            
        ax1.plot(centers, pdf)
        ax1.set_xlim(energymin, energymax)

        ax1.set_xlabel('Potential Energy [eV]')
        ax1.set_ylabel('pdf(E)')

        if matplotlib_axes is None:
            return fig

    def plot_cumulative_pdf(self, nbins=301, energymin=-15.0, energymax=15.0,
                            matplotlib_axes=None, **kwargs):
        """
        Generates a plot of the cumulative probability density function of the energy.

        Parameters
        ----------
        nbins : int, optional
            The number of histogram bins to use.  Default value is 301.
        energymin : float, optional
            The minimum energy bound to consider.  Default value is -15.0.
        energymax : float, optional
            The maximum energy bound to consider.  Default value is 15.0.
        matplotlib_axes : matplotlib.Axes.axes, optional, optional
            An existing plotting axis to add the pdf plot to.  If not given,
            a new figure object will be generated.
        **kwargs : any, optional
            Any additional key word arguments will be passed to
            matplotlib.pyplot.figure for generating a new figure object (if
            axis is not given).
        
        Returns
        -------
        matplotlib.Figure
            The generated figure.  Not returned if matplotlib_axes is given.
        """
        # Get pdf and centers in the range
        cum_pdf, centers = self.cumulative_pdf(nbins=nbins, energymin=energymin, energymax=energymax)

        # Initial plot setup and parameters
        if matplotlib_axes is None:
            fig = plt.figure(**kwargs)
            ax1 = fig.add_subplot(111)
        else:
            ax1 = matplotlib_axes
            
        ax1.plot(centers, cum_pdf)
        ax1.set_xlim(energymin, energymax)

        ax1.set_xlabel('Potential Energy [eV]')
        ax1.set_ylabel('cumulative pdf(E)')

        if matplotlib_axes is None:
            return fig


# coding: utf-8
# Standard Python libraries
from pathlib import Path
import shlex
import subprocess

# atomman imports
from . import Log, LammpsError

def run(lammps_command, script_name=None, script=None, mpi_command=None,
        restart_script_name=None, restart_script=None, logfile='log.lammps',
        screen=True, suffix=None):
    """
    Calls LAMMPS to run. Returns a Log object of the screen/logfile output.
    
    Parameters
    ----------
    lammps_command : str
        The LAMMPS inline run command (sans -in script_name).
    script_name : str, optional
        Path of the LAMMPS input script file to use.  Either script_name or
        script must be given.
    script : str, optional
        The LAMMPS input script command lines to use.  Either script_name or
        script must be given.
    mpi_command : str, optional
        The MPI inline command to run LAMMPS in parallel. Default value is 
        None (no mpi).
    restart_script_name : str, optional
        Path to an alternate LAMMPS input script file to use for restart runs.
        If given, the restart script will be used if the specified logfile
        already exists.  Requires logfile to not be None.
    restart_script : str, optional
        Alternate LAMMPS script command lines to use for restart runs.
        If given, the restart script will be used if the specified logfile
        already exists.  Requires logfile to not be None.
    logfile : str or None, optional
        Specifies the path to the logfile to write to.  Default value is
        'log.lammps'.  If set to None, then no logfile will be created.
    screen : bool, optional
        If True (default), then the resulting Log object is built from the
        LAMMPS screen output.  If False, then LAMMPS outputs no screen info
        and the Log object will be built by reading logfile.
    suffix : str, optional
        Allows for the LAMMPS suffix option to be specified to use any of the
        accelerated versions of pair styles if available.
    
    Returns
    -------
    atomman.lammps.Log 
        Contains the processed screen/logfile contents.  Will not be returned
        if logfile is None and screen is False as there would be no means to
        capture the LAMMPS output.
    """

    # Check if either restart_script_name or restart_script is given
    if restart_script_name is not None or restart_script is not None:
        if restart_script_name is not None and restart_script is not None:
            raise  ValueError('Cannot give both restart_script and restart_script_name')
        if logfile is None:
            raise ValueError('logfile must be given to automatically determine restart status')
        else:
            logfile = Path(logfile)
        
        # Check if simulation was previously started by looking for the logfile
        if logfile.is_file():
            logname = logfile.stem
            logext = logfile.suffix
            
            # Replace script parameters with restart_script parameters
            script = restart_script
            script_name = restart_script_name
            
            # Search for any earlier log files with the name log-*.lammps
            maxlogid = 0
            for oldlog in Path().glob(f'{logname}-*{logext}'):
                logid = int(oldlog.stem.split('-')[-1])
                if logid > maxlogid:
                    maxlogid = logid
            
            # Rename old logfile to keep it from being overwritten
            lognum = maxlogid + 1
            logfile.rename(f'{logname}-{lognum}{logext}')
        else:
            lognum = 0
    else:
        lognum = 0
    

    # Initialize the run command
    command = ''

    # Add mpi_command
    if mpi_command is not None:
        command += mpi_command + ' '
    
    # Add lammps_command
    command += lammps_command + ' '

    # Add logfile
    if logfile is None:
        logfile = 'none'
    if logfile != 'log.lammps':
        command += f'-log {logfile} '

    # Add script_name
    if script_name is not None:
        if script is not None:
            raise  ValueError('Cannot give both script and script_name')
        command += f'-in {script_name} '
    elif script is None:
        raise ValueError('script or script_name must be given')

    # Add screen
    if screen is False:
        command += f'-screen none '

    # Add suffix
    if suffix is not None:
        command += f'-suffix {suffix} '

    # Use shlex to split command
    command = shlex.split(command)
    
    # Try to run lammps as a subprocess
    try:
        output = subprocess.run(command, input=script, check=True,
                                capture_output=True, text=True)
    
    # Convert LAMMPS error to a Python error if failed
    except subprocess.CalledProcessError as e:
        raise LammpsError(e.output)
    
    # Initialize Log object
    log = Log()
    
    # Read in all old runs
    for i in range(1, lognum+1):
        log.read(f'{logname}-{i}{logext}')
    
    # Read in current run
    if screen:
        log.read(output.stdout)
    else:
        log.read(logfile)
    
    return log